module Error(putLogLn,putLog, attempt, tryElse, tryMost, tryMost_, catchMost, tryMap, tryMapM, first, 
    handleMost, ioElse, withErrorLog, indent, retry, putLogException, withStartEndEntrys, trySeveral, 
    emapM, eannM, withErrorMessage) where

import Control.Exception as E
import IO hiding(bracket, try, catch)
import IOExts
import Monad
import Control.Concurrent.MVar
import Control.Concurrent
import Time(getClockTime)
import List(delete)
import Charset
import EIO
import GenUtil

------------
-- Error log
------------


{-# NOINLINE ior #-}
ior :: MVar Handle
ior = unsafePerformIO $ newMVar stderr

withErrorLog :: String -> IO a -> IO a
withErrorLog fn action = E.bracket (openFile fn AppendMode) hClose $ \h -> do
	hSetBuffering h LineBuffering
	bracket (swapMVar ior h) (swapMVar ior) (\_ -> action)

withStartEndEntrys :: String -> IO a -> IO a
withStartEndEntrys n action = do
    gct >>= \ct -> putLogLn (ct ++ " " ++ n ++ " Starting")
    handle 
	(\e -> gct >>= \ct -> putLogException (ct ++ " " ++ n ++ " Ending due to Exception:" ) e >> throw e)
	(action >>= \r -> gct >>= \ct -> putLogLn (ct ++ " " ++ n ++ " Ending") >> return r) where
	    gct = getClockTime >>= \ct -> return $ "[" ++ show ct ++ "]"
    
withErrorMessage :: String -> IO a -> IO a
withErrorMessage n action = do
    handleMost 
	(\e -> gct >>= \ct -> putErrLn (normalize n ++ ct ++ " Ending due to Exception:\n" ++ indent 4 (show e) ) >> throw e )
	action  where
	    gct = getClockTime >>= \ct -> return $ "[" ++ show ct ++ "]"
    
    
    

logBracket = withMVar ior


normalize :: String -> String
normalize = unlines . lines

putLogLn :: String -> IO ()
putLogLn s = putLog (s ++ "\n")

-- | log entry
putLog :: String -> IO ()
putLog s = logBracket (\h -> putRaw h (stringToBytes s))


-- | transform an exception with a function.
emapM :: (Exception -> Exception) -> IO a -> IO a
emapM f action = do
    handle (\e -> ioError (f e)) action  

eannM :: String -> IO a -> IO a
eannM s action = emapM f action where
    f (ErrorCall es) = ErrorCall $ normalize s ++ normalize es
    f e = ErrorCall $ normalize s ++ normalize (show e)

attempt :: IO a -> IO ()
attempt action = tryMost action >>= \x -> case x of 
    Left e -> putLogException "attempt ExceptionCaught" e 
    Right _ -> return ()


tryElse r x = tryMost x >>= \y -> case y of
    Left e -> putLogException "tryElse ExceptionCaught" e >> return r
    Right v -> return v

tryMost_ x = tryMost x >> return ()

tryMap :: (a -> b) -> [a] -> IO [b]
tryMap f xs = do 
    ys <- mapM (tryMost . evaluate . f ) xs
    return [y|(Right y) <- ys]
    
tryMapM :: (a -> IO b) -> [a] -> IO [b]
tryMapM f xs = do 
    ys <- mapM (tryMost . f ) xs
    return [y|(Right y) <- ys]

tryMost = E.tryJust passKilled

passKilled (AsyncException ThreadKilled) = Nothing
passKilled x = Just x

catchMost = E.catchJust passKilled
handleMost = E.handleJust passKilled

first :: [IO a] -> IO a
first [] = fail "empty  argument to first"
first [x] = x
first (x:xs) = E.try x >>= \z -> case z of
    Left e@(AsyncException ThreadKilled) -> throw e
    Left _ -> first xs
    Right v -> return v

ioElse :: IO a -> IO a -> IO a
ioElse a b = tryMost a >>= \x -> case x of 
    Left _ -> b
    Right x -> return x

indent :: Int -> String -> String 
indent n s = unlines $ map (replicate n ' ' ++)$ lines s 

retry :: Float -> String -> IO a -> IO a
retry delay n action = do
    handleMost (\e -> putLogException (n ++ " (retrying in " ++ show delay ++ "s):") e >> threadDelay (floor $ 1000000 * delay) >> retry delay n action) action


putLogException :: String -> Exception -> IO ()
putLogException n e =  putLog (n ++ "\n" ++ indent 4 (show e)) 


-- concurrently try several IO actions, returning the result of the first to finish.
-- if all actions throw exceptions, one is passed on non-deterministically
trySeveral :: [IO a] -> IO a
trySeveral [] = error "trySeveral has nothing to try!"
trySeveral arms = do
    v <- newEmptyMVar 
    ts <- mapM (forkIO . f v) arms
    g v ts where
	f v arm = do
	    t <- myThreadId
	    r <- tryMost arm
	    putMVar v (t,r)
	g v ts = do
	    (t,r) <- takeMVar v
	    let ts' = delete t ts
	    case r of 
		Left e -> if null ts' then throw e else g v ts'
		Right x -> do
		    mapM_ killThread ts'
		    return x
